"""In this example, the automaton neighbours are a orthogonal rectangle with
the following features:
 - One neighbours dimension is longer than the other
 - One neighbours dimension is pair (no symmetric)
"""
import ndautomata as ca
import matplotlib.animation as animation
import matplotlib.pyplot as plt
from ndautomata import initializers
from ndautomata import neighbours


# -------------------------------------------------------------------
# Automaton ---------------------------------------------------------
# Defined by dimensions, neighbours and states
class Automaton(ca.BaseAutomaton):
    neighbours = neighbours.orthogonal([5, 2])
    states = 2


# -------------------------------------------------------------------
# Rule --------------------------------------------------------------
# Requires the same shape as the automaton neighbours
# Dimensions size equals the automaton states
connections = [Automaton.states] * Automaton.neighbours.size
rule = initializers.random(states=Automaton.states, size=connections)


# -------------------------------------------------------------------
# Automaton instantiation -------------------------------------------
# Initialization requires states and size
ic = initializers.center(states=Automaton.states, size=[100, 100])
automaton = Automaton(ic, rule)


# -------------------------------------------------------------------
# Evolution and plot ------------------------------------------------
# Iterate using next and list comprehensions
plt.rcParams["image.cmap"] = "binary"
fig = plt.figure()
plt.matshow(ic, fignum=0)
updt = lambda *args: plt.matshow(next(automaton), fignum=0)  # noqa: E731
anim = animation.FuncAnimation(fig, updt, frames=200, repeat=False)
plt.show()



"""Module to test automaton features and requirements.
Note order of bits in rules is from lowest to biggest.
"""
import numpy as np
from ndautomata import BaseAutomaton, neighbours, initializers
from pytest import mark


def srule(buffer):
    array = np.frombuffer(buffer, dtype="uint8")[::-1]
    return np.unpackbits(array, bitorder="little")


class TestElementaryR1S2:
    class Automaton(BaseAutomaton):
        neighbours = neighbours.regular(ndim=1, r=1)
        states = 2

    ic = np.array([0, 1, 0, 0, 1, 1, 0, 1, 0, 0], dtype="uint8")

    @mark.parametrize("repeat", range(10))
    def test_rule_indexes(self, repeat):
        rule = initializers.random(states=2, size=[2] * 3)
        nc = next(self.Automaton(self.ic, rule))
        assert nc[0] == rule[0, 0, 1]
        assert nc[1] == rule[0, 1, 0]
        assert nc[2] == rule[1, 0, 0]
        assert nc[3] == rule[0, 0, 1]
        assert nc[4] == rule[0, 1, 1]
        assert nc[5] == rule[1, 1, 0]
        assert nc[6] == rule[1, 0, 1]
        assert nc[7] == rule[0, 1, 0]
        assert nc[8] == rule[1, 0, 0]
        assert nc[9] == rule[0, 0, 0]

    def test_rule_0b00000000(self):  # 000
        rule = srule(b"\x00").reshape(2, 2, 2)
        nc = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b00000001(self):  # 001
        rule = srule(b"\x01").reshape(2, 2, 2)
        nc = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b00100000(self):  # 032
        rule = srule(b"\x20").reshape(2, 2, 2)
        nc = np.array([0, 0, 0, 0, 0, 0, 1, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b01101110(self):  # 110
        rule = srule(b"\x6e").reshape(2, 2, 2)
        nc = np.array([1, 1, 0, 1, 1, 1, 1, 1, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b01111101(self):  # 125
        rule = srule(b"\x7d").reshape(2, 2, 2)
        nc = np.array([0, 1, 1, 0, 1, 1, 1, 1, 1, 1], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b11010111(self):  # 215
        rule = srule(b"\xd7").reshape(2, 2, 2)
        nc = np.array([1, 1, 1, 1, 0, 1, 0, 1, 1, 1], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b11111110(self):  # 254
        rule = srule(b"\xfe").reshape(2, 2, 2)
        nc = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0b11111111(self):  # 255
        rule = srule(b"\xff").reshape(2, 2, 2)
        nc = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))


class TestElementaryR2S2:
    class Automaton(BaseAutomaton):
        neighbours = neighbours.regular(ndim=1, r=2)
        states = 2

    ic = np.array([0, 1, 0, 0, 1, 1, 0, 1, 0, 0], dtype="uint8")

    @mark.parametrize("repeat", range(10))
    def test_rule_indexes(self, repeat):
        rule = initializers.random(states=2, size=[2] * 5)
        nc = next(self.Automaton(self.ic, rule))
        assert nc[0] == rule[0, 0, 0, 1, 0]
        assert nc[1] == rule[0, 0, 1, 0, 0]
        assert nc[2] == rule[0, 1, 0, 0, 1]
        assert nc[3] == rule[1, 0, 0, 1, 1]
        assert nc[4] == rule[0, 0, 1, 1, 0]
        assert nc[5] == rule[0, 1, 1, 0, 1]
        assert nc[6] == rule[1, 1, 0, 1, 0]
        assert nc[7] == rule[1, 0, 1, 0, 0]
        assert nc[8] == rule[0, 1, 0, 0, 0]
        assert nc[9] == rule[1, 0, 0, 0, 1]

    def test_rule_0x00000000(self):
        rule = srule(b"\x00\x00\x00\x00").reshape([2] * 5)
        nc = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0x00000001(self):
        rule = srule(b"\x00\x00\x00\x01").reshape([2] * 5)
        nc = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0x00000004(self):
        rule = srule(b"\x00\x00\x00\x04").reshape([2] * 5)
        nc = np.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0x00000200(self):
        rule = srule(b"\x00\x00\x02\x00").reshape([2] * 5)
        nc = np.array([0, 0, 1, 0, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0x00000204(self):
        rule = srule(b"\x00\x00\x02\x04").reshape([2] * 5)
        nc = np.array([1, 0, 1, 0, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0x00080000(self):
        rule = srule(b"\x00\x08\x00\x00").reshape([2] * 5)
        nc = np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0x00080204(self):
        rule = srule(b"\x00\x08\x02\x04").reshape([2] * 5)
        nc = np.array([1, 0, 1, 1, 0, 0, 0, 0, 0, 0], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0xffffffff(self):
        rule = srule(b"\xff\xff\xff\xff").reshape([2] * 5)
        nc = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))


class TestElementaryR2S3:
    class Automaton(BaseAutomaton):
        neighbours = neighbours.regular(ndim=1, r=2)
        states = 3

    ic = np.array([0, 1, 2, 0, 1, 1, 0, 1, 2, 0], dtype="uint8")

    @mark.parametrize("repeat", range(10))
    def test_rule_indexes(self, repeat):
        rule = initializers.random(states=3, size=[3] * 5)
        nc = next(self.Automaton(self.ic, rule))
        assert nc[0] == rule[2, 0, 0, 1, 2]
        assert nc[1] == rule[0, 0, 1, 2, 0]
        assert nc[2] == rule[0, 1, 2, 0, 1]
        assert nc[3] == rule[1, 2, 0, 1, 1]
        assert nc[4] == rule[2, 0, 1, 1, 0]
        assert nc[5] == rule[0, 1, 1, 0, 1]
        assert nc[6] == rule[1, 1, 0, 1, 2]
        assert nc[7] == rule[1, 0, 1, 2, 0]
        assert nc[8] == rule[0, 1, 2, 0, 0]
        assert nc[9] == rule[1, 2, 0, 0, 1]


class TestOrthogonalR1S2:
    class Automaton(BaseAutomaton):
        neighbours = neighbours.regular(ndim=2, r=1)
        states = 2

    ic = np.array([[0, 1, 0, 0, 1, 1, 0, 1, 0, 0]] * 6, dtype="uint8")

    @mark.parametrize("repeat", range(10))
    def test_rule_indexes(self, repeat):
        rule = initializers.random(states=2, size=[2] * 9)
        nc = next(self.Automaton(self.ic, rule))
        assert nc[0][0] == rule[0, 0, 1, 0, 0, 1, 0, 0, 1]
        assert nc[0][1] == rule[0, 1, 0, 0, 1, 0, 0, 1, 0]
        assert nc[0][2] == rule[1, 0, 0, 1, 0, 0, 1, 0, 0]
        assert nc[0][3] == rule[0, 0, 1, 0, 0, 1, 0, 0, 1]
        assert nc[0][4] == rule[0, 1, 1, 0, 1, 1, 0, 1, 1]
        assert nc[0][5] == rule[1, 1, 0, 1, 1, 0, 1, 1, 0]
        assert nc[0][6] == rule[1, 0, 1, 1, 0, 1, 1, 0, 1]
        assert nc[0][7] == rule[0, 1, 0, 0, 1, 0, 0, 1, 0]
        assert nc[0][8] == rule[1, 0, 0, 1, 0, 0, 1, 0, 0]
        assert nc[0][9] == rule[0, 0, 0, 0, 0, 0, 0, 0, 0]

    def test_rule_0_0(self):
        rule = np.array([0] * 2 ** 9).reshape([2] * 9)
        nc = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0]] * 6, dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_0_1(self):
        rule = np.array([1] + [0] * (2 ** 9 - 1)).reshape([2] * 9)
        nc = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 1]] * 6, dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))

    def test_rule_1_1(self):
        rule = np.array([1] * 2 ** 9).reshape([2] * 9)
        nc = np.array([[1, 1, 1, 1, 1, 1, 1, 1, 1, 1]] * 6, dtype="uint8")
        assert np.all(nc == next(self.Automaton(self.ic, rule)))



"""Module to test neural automaton evolution operations"""
from ndautomata import BaseAutomaton, neighbours, initializers
from pytest import fixture


@fixture(scope="session", params=[1, 2])
def ndim(request):
    return request.param


@fixture(scope="session", params=[1])
def radius(request):
    return request.param


@fixture(scope="session", params=[2, 3])
def nstates(request):
    return request.param


@fixture(scope="session")
def automaton_class(ndim, radius, nstates):
    class Automaton(BaseAutomaton):
        neighbours = neighbours.regular(ndim, radius)
        states = nstates

    return Automaton


@fixture(scope="session", params=[10, 20])
def shape(request, ndim):
    return [request.param] * ndim


@fixture(scope="session", params=[initializers.random])
def rule(request, nstates, automaton_class):
    connections = [nstates] * automaton_class.neighbours.size
    return request.param(states=nstates, size=connections)



from ndautomata import neighbours
import numpy as np


def test_1dr1():
    expected = np.flip(np.arange(3).reshape((3,)) + 1)
    assert np.all(neighbours.regular(ndim=1, r=1) == expected)


def test_1dr2():
    expected = np.flip(np.arange(5).reshape((5,)) + 1)
    assert np.all(neighbours.regular(ndim=1, r=2) == expected)


def test_2dr1():
    expected = np.flip(np.arange(9).reshape((3, 3)) + 1)
    assert np.all(neighbours.regular(ndim=2, r=1) == expected)


def test_2dr2():
    expected = np.flip(np.arange(25).reshape((5, 5)) + 1)
    assert np.all(neighbours.regular(ndim=2, r=2) == expected)


def test_3dr1():
    expected = np.flip(np.arange(27).reshape((3, 3, 3)) + 1)
    assert np.all(neighbours.regular(ndim=3, r=1) == expected)


def test_3dr2():
    expected = np.flip(np.arange(125).reshape((5, 5, 5)) + 1)
    assert np.all(neighbours.regular(ndim=3, r=2) == expected)



"""The Game of Life, also known simply as Life, is a cellular automaton
devised by the British mathematician John Horton Conway in 1970. It is a
zero-player game, meaning that its evolution is determined by its initial
state, requiring no further input. One interacts with the Game of Life by
creating an initial configuration and observing how it evolves. It is Turing
complete and can simulate a universal constructor.
"""
import ndautomata as ca
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
from ndautomata import initializers
from ndautomata import neighbours


# -------------------------------------------------------------------
# Automaton ---------------------------------------------------------
# Defined by dimensions, neighbours and states
class Automaton(ca.BaseAutomaton):
    neighbours = neighbours.regular(ndim=2, r=1)
    states = 2


# -------------------------------------------------------------------
# Rule --------------------------------------------------------------
# Requires the same shape as the automaton neighbours
# Dimensions size equals the automaton states
connections = [Automaton.states] * Automaton.neighbours.size
rule = initializers.zeros(states=Automaton.states, size=connections)
for index in np.ndindex(rule.shape):
    match index[4]:  # State
        # Any alive cell that is touching less than two alive neighbours dies
        case 1 if sum(index) <= 2:
            rule[index] = 0  # dies
        # Any alive cell touching four or more alive neighbours dies
        case 1 if sum(index) >= 5:
            rule[index] = 0  # dies
        # Any dead cell touching exactly three alive neighbours becomes alive
        case 0 if sum(index) == 3:
            rule[index] = 1  # lives
        # Any alive cell touching two or three alive neighbours does nothing
        case state:
            rule[index] = state


# -------------------------------------------------------------------
# Automaton instantiation -------------------------------------------
# Initialization requires states and size
ic = initializers.random(states=2, size=[100, 100])
automaton = Automaton(ic, rule)


# -------------------------------------------------------------------
# Evolution and plot ------------------------------------------------
# Iterate using next and list comprehensions
plt.rcParams["image.cmap"] = "binary"
fig = plt.figure()
plt.matshow(ic, fignum=0)
updt = lambda *args: plt.matshow(next(automaton), fignum=0)  # noqa: E731
anim = animation.FuncAnimation(fig, updt, frames=200, repeat=False)
plt.show()



"""Module for initializer functions."""
import numpy as np


def zeros(states, size):
    """Returns an array where all values are zero.
    :param states: Number of possible states, not used
    :param size: Shape of the return array
    :return: numpy ndarray
    """
    return np.zeros(size, dtype="uint8")


def ones(states, size):
    """Returns an array where all values are one.
    :param states: Number of possible states, not used
    :param size: Shape of the return array
    :return: numpy ndarray
    """
    return (states - 1) * np.ones(size, dtype="uint8")


def center(states, size):
    """Returns an array where all values are zero except the center.
    :param states: Number of possible states, not used
    :param size: Shape of the return array
    :return: numpy ndarray
    """
    middle = tuple(int(d / 2.0) for d in size)
    values = zeros(states, size)
    values[middle] = states - 1
    return values


def border(states, size):
    """Returns an array where all values are zero except the border.
    :param states: Number of possible states, not used
    :param size: Shape of the return array
    :return: numpy ndarray
    """
    border = tuple(0 for _ in size)
    values = zeros(states, size)
    values[border] = states - 1
    return values


def random(states, size):
    """Returns an array where all values are random.
    :param states: Number of possible states, not used
    :param size: Shape of the return array
    :return: numpy ndarray
    """
    return np.random.randint(states, size=size, dtype="uint8")



"""Automata based on bidimensional grid of square cells, each of which is in
one of two possible states. Every cell interacts only with six neighbours,
which are the cells that are horizontally, vertically, or half diagonally
adjacent.                         _________________
         / \ / \ / \             |     |     |     |
        |0,0|0,1|0,2|            |     |-1,0 |-1,+1|
         \ / \ / \ / \           |-----|-----|-----|
          |1,0|1,1|1,2|          | 0,-1| i,j | 0,+1|
           \ / \ / \ / \         |-----|-----|-----|
            |2,0|2,1|2,2|        |+1,-1|+1,0 |     |
             \ / \ / \ /         |_____|_____|_____|
"""  # noqa: W605
import ndautomata as ca
import matplotlib.animation as animation
import matplotlib.pyplot as plt
from ndautomata import initializers
from ndautomata import neighbours


# -------------------------------------------------------------------
# Automaton ---------------------------------------------------------
# Defined by dimensions, neighbours and states
class Automaton(ca.BaseAutomaton):
    neighbours = neighbours.hexagonal(ndim=2, r=1)
    states = 2


# -------------------------------------------------------------------
# Rule --------------------------------------------------------------
# Requires the same shape as the automaton neighbours
# Dimensions size equals the automaton states
connections = [Automaton.states] * Automaton.neighbours.size
rule = initializers.random(states=Automaton.states, size=connections)

# -------------------------------------------------------------------
# Automaton instantiation -------------------------------------------
# Initialization requires states and size
ic = initializers.center(states=2, size=[100, 100])
automaton = Automaton(ic, rule)


# -------------------------------------------------------------------
# Evolution and plot ------------------------------------------------
# Iterate using next and list comprehensions
# TODO: improve using matplotlib.collections.RegularPolyCollection
plt.rcParams["image.cmap"] = "binary"
fig = plt.figure()
plt.matshow(ic, fignum=0)
updt = lambda *args: plt.matshow(next(automaton), fignum=0)  # noqa: E731
anim = animation.FuncAnimation(fig, updt, frames=200, repeat=False)
plt.show()



"""Automata as three-dimensional orthogonal grid of cubic cells, each of
which is in one of two possible states. Every cell interacts with its twenty
seven neighbours, which are the cells that are horizontally, vertically or
diagonally adjacent."""
import ndautomata as ca
import matplotlib.pyplot as plt
import numpy as np
from ndautomata import initializers
from ndautomata import neighbours


# -------------------------------------------------------------------
# Automaton ---------------------------------------------------------
# Defined by dimensions, neighbours and states
class Automaton(ca.BaseAutomaton):
    neighbours = neighbours.regular(ndim=3, r=1)
    states = 2


# -------------------------------------------------------------------
# Rule --------------------------------------------------------------
# Requires the same shape as the automaton neighbours
# Dimensions size equals the automaton states
connections = [Automaton.states] * Automaton.neighbours.size
rule = initializers.random(states=Automaton.states, size=connections)
rule[1::3] = 0  # Constrain for the rule


# -------------------------------------------------------------------
# Automaton instantiation -------------------------------------------
# Initialization requires states and size
ic = initializers.center(states=Automaton.states, size=[40, 40, 40])
automaton = Automaton(ic, rule)


# -------------------------------------------------------------------
# Evolution and plot ------------------------------------------------
# Iterate using next and mean using np.mean
plt.rcParams["image.cmap"] = "binary"
data = np.array([next(automaton) for _ in range(100)])
fig, axs = plt.subplots(1, automaton.dimensions)
for dim in range(automaton.dimensions):
    axis = tuple(d + 1 for d in range(automaton.dimensions) if d != dim)
    axs[dim].matshow(np.mean(data, axis=axis))
plt.show()



from ndautomata import BaseAutomaton
from pytest import fixture


@fixture(scope="module")
def example():
    import examples.elementary as example

    return example


def test_automaton(example):
    assert isinstance(example.automaton, BaseAutomaton)



"""Module to test automaton features and requirements"""
import copy

import numpy as np
from ndautomata import initializers
from pytest import fixture, mark


# Module fixtures ---------------------------------------------------
@fixture(scope="class", params=[initializers.random])
def initial_configuration(request, nstates, shape):
    return request.param(nstates, shape)


@fixture(scope="class")
def ic(initial_configuration):
    return initial_configuration


@fixture(scope="class")
def automaton(automaton_class, initial_configuration, rule):
    return automaton_class(initial_configuration, rule)


# Requirements ------------------------------------------------------
class AttrRequirements:
    def test_attr_rule(self, automaton):
        assert hasattr(automaton, "rule")
        assert isinstance(automaton.rule, np.ndarray)

    def test_attr_configuration(self, automaton):
        assert hasattr(automaton, "configuration")
        assert isinstance(automaton.configuration, np.ndarray)

    def test_method_next(self, automaton):
        assert hasattr(automaton, "__next__")
        assert callable(automaton.__next__)

    def test_ic_memory(self, automaton, ic):
        assert not np.shares_memory(automaton.configuration, ic)


class NextRequirements:
    def test_returns_ndarray(self, next):
        assert isinstance(next, np.ndarray)

    def test_keep_rule(self, original, rule):
        assert np.all(original.rule == rule)

    @mark.parametrize("rule", [initializers.zeros], indirect=True)
    def test_indexing_rule0(self, next):
        assert np.all(next == 0)

    @mark.parametrize("rule", [initializers.ones], indirect=True)
    def test_indexing_rule1(self, next, nstates):
        assert np.all(next == nstates - 1)

    def test_next_copies(self, automaton, next):
        assert automaton.configuration is not next
        assert np.all(automaton.configuration == next)


# Parametrization ---------------------------------------------------
class TestAutomaton(AttrRequirements, NextRequirements):
    @fixture(scope="function")
    def original(self, automaton):
        return copy.deepcopy(automaton)

    @fixture(scope="function")
    def next(self, automaton, original):
        return next(automaton)



"""A cellular automaton consists of a regular grid of cells, each in
one of a finite number of states. The grid can be in any finite number
of dimensions. For each cell, a set of cells called its neighborhood
is defined relative to the specified cell.

An initial state (time t = 0) is selected by assigning a state for
each cell. A new generation is created (advancing t by 1), according
to a rule function that determines the new state of each cell in terms
of the current state of the cell and the states of the cells in its
neighborhood.
"""
import copy
from abc import ABC

import numpy as np
from pydantic import PositiveInt
from scipy.ndimage import correlate

from ndautomata import neighbours


class BaseAutomaton(ABC):
    """Abstract class for automaton generation. You can generate your own
    automaton by subclassing this class defining your own attributes.

    Parameters
    ----------
    initial_configuration : (N,) ndarray
        Initial configuration for all the automaton cells.
        Cell values must be equal or lower than the number of states.
        `all(CA.configuration <= Automaton.states)`
    rule :  (N,) ndarray
        Indexing for next cell value following neighborhood states.
        Rule dimensions mut match the neighbours size.
        Dimension length must be equal or lower than the number of states.
        `Rule.shape ~= [Automaton.states] * Automaton.neighbours.size`

    Attributes
    ----------
    configuration : (N,) ndarray
        NdArray containing all current cell states for the cellular automaton.
    dimensions : PositiveInt
        Number of dimensions the cells are arranged in the cellular automaton.
    rule : (N,) ndarray
        Rule used to calculate next cell states in the cellular automaton.

    Class Attributes
    ----------------
    neighbours :  (N,) ndarray
        Relative indexing for each cell in the cellular automaton.
    states : PositiveInt
        Amount of possible states a cell can take.
    """

    neighbours: np.ndarray
    states: PositiveInt

    def __init__(self, initial_configuration, rule):
        assert initial_configuration.ndim == self.dimensions
        assert np.max(initial_configuration) < self.states
        self.configuration = copy.copy(initial_configuration)
        self.rule = rule
        self.__index = np.empty(initial_configuration.shape, dtype="uint")
        self._weights = np.array(self.states**self.neighbours, dtype="uint")
        self._weights //= self.states

    def __next__(self):
        correlate(
            self.configuration,  # Automaton states and neighbours
            self._weights,  # Correlation with connection weights
            output=self.__index,  # Output should be uint max
            mode="wrap",  # ‘wrap’ (a b c d | a b c d | a b c d)
        )
        self.configuration[:] = self._rule[self.__index]
        return copy.deepcopy(self.configuration)

    @classmethod
    @property
    def dimensions(cls):
        return cls.neighbours.ndim

    @property
    def rule(self):
        rule_shape = [self.states] * self.neighbours.size
        return self._rule.reshape(rule_shape)

    @rule.setter
    def rule(self, value):
        assert isinstance(value, np.ndarray)
        assert len(value.shape) == self.neighbours.size
        assert np.max(value) < self.states
        self._rule = value.ravel()


if __name__ == "__main__":
    from timeit import timeit

    from ndautomata import initializers

    class Automaton(BaseAutomaton):
        dimensions = 3
        neighbours = neighbours.regular(ndim=3, r=1)
        states = 2

    # Configuration
    connections = [Automaton.states] * Automaton.neighbours.size
    rule = initializers.random(states=Automaton.states, size=connections)
    ic = initializers.random(states=Automaton.states, size=[100, 100, 10])
    ca = Automaton(ic, rule)

    # Profiling
    time = timeit(lambda: [next(ca) for _ in range(100)], number=10)
    print(f"cc next timing: {time}")



"""In mathematics and computability theory, an elementary cellular automaton
is a one-dimensional cellular automaton where there are two possible states
(labeled 0 and 1) and the rule to determine the state of a cell in the next
generation depends only on the current state of the cell and its two immediate
neighbors.
"""
import ndautomata as ca
import matplotlib.pyplot as plt
import numpy as np
from ndautomata import initializers
from ndautomata import neighbours


# -------------------------------------------------------------------
# Automaton ---------------------------------------------------------
# Defined by dimensions, neighbours and states
class Automaton(ca.BaseAutomaton):
    neighbours = neighbours.regular(ndim=1, r=1)
    states = 2


# -------------------------------------------------------------------
# Rule --------------------------------------------------------------
# Requires the same shape as the automaton neighbours
# Dimensions size equals the automaton states
connections = [Automaton.states] * Automaton.neighbours.size
rule = np.unpackbits(np.array([110], dtype="uint8"), bitorder="little")
rule = rule.reshape(connections)


# -------------------------------------------------------------------
# Automaton instantiation -------------------------------------------
# Initialization requires states and size
ic = initializers.center(states=Automaton.states, size=[100])
automaton = Automaton(ic, rule)


# -------------------------------------------------------------------
# Evolution and plot ------------------------------------------------
# Iterate using next and list comprehensions
plt.rcParams["image.cmap"] = "binary"
plt.matshow([next(automaton) for _ in range(200)])
plt.show()
